import concurrent.futures
from typing import Any, Callable, Iterable, List, Optional


def concurrent_forloop(
    func: Callable[..., Any],
    iterable: Iterable,
    *iterables: Iterable,
    max_workers: Optional[int] = None,
) -> List[Any]:
    """
    Run a function concurrently on an iterable and return the results in order.

    If max_workers is None, the ThreadPoolExecutor will determine the optimal
    number of workers based on system resources.

    This function also supports passing additional iterables to provide varying arguments
    to the function. When additional iterables are provided, the function is called
    with one item from each iterable concurrently, similar to the built-in map function.

    Args:
        func (Callable): The function to run on each element of the iterable.
        iterable (Iterable): The primary iterable to process.
        *iterables (Iterable): Optional additional iterables that provide varying arguments.
        max_workers (Optional[int]): The maximum number of workers to use. Defaults to None,
                                     which lets ThreadPoolExecutor decide based on system
                                     resources.

    Returns:
        List[Any]: A list of results from the function.
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        return list(executor.map(func, iterable, *iterables))



# poted from controlnet repo
from typing import List, NamedTuple, Optional


class Keypoint(NamedTuple):
    x: float
    y: float
    score: float = 1.0
    id: int = -1


class BodyResult(NamedTuple):
    # Note: Using `Optional` instead of `|` operator as the ladder is a Python
    # 3.10 feature.
    # Annotator code should be Python 3.8 Compatible, as controlnet repo uses
    # Python 3.8 environment.
    # https://github.com/lllyasviel/ControlNet/blob/d3284fcd0972c510635a4f5abe2eeb71dc0de524/environment.yaml#L6
    keypoints: List[Optional[Keypoint]]
    total_score: float = 0.0
    total_parts: int = 0


HandResult = List[Keypoint]
FaceResult = List[Keypoint]


class PoseResult(NamedTuple):
    body: BodyResult
    left_hand: Optional[HandResult]
    right_hand: Optional[HandResult]
    face: Optional[FaceResult]



import hashlib
import os
import re
import shutil
import sys
import tempfile
import zipfile
from glob import glob
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse  # noqa: F401
from urllib.request import Request, urlopen

from tqdm import tqdm


def _get_cache_dir():
    cache_dir = os.path.expanduser("~/.cache/spinepose")
    return os.path.join(cache_dir, "hub")


def extract_zip(zip_file_path, extract_to_path):
    if not os.path.exists(extract_to_path):
        os.makedirs(extract_to_path)

    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
        zip_ref.extractall(extract_to_path)


def download_url_to_file(url, dst, hash_prefix=None, progress=True):
    """Download object at the given URL to a local path.

    Modified from `torch.hub.download_url_to_file`.

    Args:
        url (str): URL of the object to download
        dst (str): Full path where object will be saved, for example,
            ``/tmp/temporary_file``.
        hash_prefix (str, optional): If not None, the SHA256 downloaded
            file should start with ``hash_prefix``. Defaults to None.
        progress (bool): whether or not to display a progress
            bar to stderr Defaults to True.
    """
    file_size = None
    req = Request(url, headers={"User-Agent": "mmlmtools"})
    u = urlopen(req)
    meta = u.info()
    if hasattr(meta, "getheaders"):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])

    # We deliberately save it in a temp file and move it after
    # download is complete. This prevents a local working checkpoint
    # being overridden by a broken download.
    dst = Path(dst).expanduser().absolute()
    try:
        with tempfile.NamedTemporaryFile(delete=False, dir=dst.parent) as f:
            if hash_prefix is not None:
                sha256 = hashlib.sha256()
            with tqdm(
                total=file_size,
                disable=not progress,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as pbar:
                while True:
                    buffer = u.read(8192)
                    if len(buffer) == 0:
                        break
                    f.write(buffer)
                    if hash_prefix is not None:
                        sha256.update(buffer)
                    pbar.update(len(buffer))

        if hash_prefix is not None:
            digest = sha256.hexdigest()
            if digest[: len(hash_prefix)] != hash_prefix:
                raise RuntimeError(
                    'invalid hash value (expected "{}", got "{}")'.format(
                        hash_prefix, digest
                    )
                )
        Path(f.name).rename(dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def download_checkpoint(
    url: str,
    dst_dir: Optional[str] = None,
    progress: bool = True,
    check_hash: bool = False,
    filename: Optional[str] = None,
) -> str:
    """Download the checkpoint from the given URL.

    Modified from `torch.hub.load_state_dict_from_url`.

    If the object is already present in `dst_dir`, it will be returned
    directly.
    The default value of ``dst_dir`` is the same as the checkpoint cache
    path of PyTorch hub.

    Args:
        url (str): URL of the object to download
        model_dir (str, optional): directory in which to save the object
        progress (bool, optional): whether or not to display a progress bar to
            stderr. Defaults to True.
        check_hash(bool, optional): If True, the filename part of the URL
            should follow the naming convention ``filename-<sha256>.ext`` where
            ``<sha256>`` is the first eight or more digits of the SHA256 hash
            of the contents of the file. The hash is used to ensure unique
            names and to verify the contents of the file. Defaults to False.
        filename (str, optional): name for the downloaded file.
            Filename from ``url`` will be used if not set.

    Returns:
        str: The path of the downloaded file.
    """
    if dst_dir is None:
        dst_dir = os.path.join(_get_cache_dir(), "checkpoints")

    dst_dir = Path(dst_dir)
    dst_dir.mkdir(parents=True, exist_ok=True)

    parts = urlparse(url)
    filename = filename or os.path.basename(parts.path)
    cached_file = dst_dir / filename
    onnx_name = Path(dst_dir, str(filename).split(".")[0] + ".onnx")

    if not cached_file.exists():
        if os.path.exists(onnx_name):
            return str(onnx_name)

        sys.stderr.write('Downloading: "{}" to {}\n'.format(url, cached_file))
        hash_prefix = None
        if check_hash:
            HASH_REGEX = re.compile(r"-([a-f0-9]*)\.")
            r = HASH_REGEX.search(filename)  # r is Optional[Match[str]]
            hash_prefix = r.group(1) if r else None
        download_url_to_file(url, cached_file, hash_prefix, progress=progress)

    if str(cached_file).split(".")[-1] == "zip":
        # os.system(f'unzip -d {dst_dir}/tmp {cached_file}')
        tmp_dir = Path(Path(cached_file).parent, "tmp")
        extract_zip(cached_file, tmp_dir)
        cached_list = glob(f"{dst_dir}/**", recursive=True)

        for each in cached_list:
            if each[-12:] == "end2end.onnx":
                cached_onnx = each
                break
        # os.system(f'mv {cached_onnx} {onnx_name}')
        # os.system(f'rm -rf {cached_file}')
        # os.system(f'rm -rf {dst_dir}/tmp')
        shutil.move(cached_onnx, onnx_name)
        os.remove(cached_file)
        shutil.rmtree(tmp_dir)

        cached_file = onnx_name

    return str(cached_file)



import cv2


def draw_bbox(img, bboxes, color=(0, 255, 0)):
    for bbox in bboxes:
        img = cv2.rectangle(
            img, (int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])), color, 2
        )
    return img


def draw_skeleton(
    img, keypoints, scores, metainfo, kpt_thr=0.5, radius=2, line_width=2
):
    keypoint_info = metainfo["keypoint_info"]
    skeleton_info = metainfo["skeleton_info"]

    if len(keypoints.shape) == 2:
        keypoints = keypoints[None, :, :]
        scores = scores[None, :, :]

    num_instance = keypoints.shape[0]
    for i in range(num_instance):
        img = draw_mmpose(
            img,
            keypoints[i],
            scores[i],
            keypoint_info,
            skeleton_info,
            kpt_thr,
            radius,
            line_width,
        )
    return img


def draw_mmpose(
    img,
    keypoints,
    scores,
    keypoint_info,
    skeleton_info,
    kpt_thr=0.5,
    radius=2,
    line_width=2,
):
    assert len(keypoints.shape) == 2

    vis_kpt = [s >= kpt_thr for s in scores]

    link_dict = {}
    for i, kpt_info in keypoint_info.items():
        link_dict[kpt_info["name"]] = kpt_info["id"]

    for i, ske_info in skeleton_info.items():
        link = ske_info["link"]
        link_color = ske_info["color"]
        pt0, pt1 = link_dict[link[0]], link_dict[link[1]]

        if vis_kpt[pt0] and vis_kpt[pt1]:
            kpt0 = keypoints[pt0]
            kpt1 = keypoints[pt1]

            img = cv2.line(
                img,
                (int(kpt0[0]), int(kpt0[1])),
                (int(kpt1[0]), int(kpt1[1])),
                link_color[::-1],
                thickness=line_width,
            )

    stroke_color = (255, 255, 255)  # White
    for i, kpt_info in keypoint_info.items():
        kpt = keypoints[i]
        fill_color = kpt_info["color"]

        if vis_kpt[i]:
            center = (int(kpt[0]), int(kpt[1]))

            # Draw outer circle (stroke)
            img = cv2.circle(
                img, center, int(radius + line_width), stroke_color[::-1], -1
            )

            # Draw inner circle (fill)
            img = cv2.circle(img, center, int(radius), fill_color[::-1], -1)

    return img



metainfo = dict(
    dataset_name="spinetrack",
    paper_info=dict(
        author="Khan, Muhammad Saif Ullah and Krauß, Stephan and Stricker, Didier",
        title="Towards Unconstrained 2D Pose Estimation of the Human Spine",
        container="CVPRW",
        year="2025",
        homepage="https://github.com/saifkhichi96/spinepose",
    ),
    keypoint_info={
        0: dict(name="nose", id=0, color=[204, 51, 51], type="upper", swap=""),
        1: dict(name="left_eye", id=1, color=[204, 51, 51], type="upper", swap="right_eye"),
        2: dict(name="right_eye", id=2, color=[204, 51, 51], type="upper", swap="left_eye"),
        3: dict(name="left_ear", id=3, color=[204, 51, 51], type="upper", swap="right_ear"),
        4: dict(name="right_ear", id=4, color=[204, 51, 51], type="upper", swap="left_ear"),
        5: dict(name="left_shoulder", id=5, color=[204, 51, 51], type="upper", swap="right_shoulder"),
        6: dict(name="right_shoulder", id=6, color=[204, 51, 51], type="upper", swap="left_shoulder"),
        7: dict(name="left_elbow", id=7, color=[204, 51, 51], type="upper", swap="right_elbow"),
        8: dict(name="right_elbow", id=8, color=[204, 51, 51], type="upper", swap="left_elbow"),
        9: dict(name="left_wrist", id=9, color=[204, 51, 51], type="upper", swap="right_wrist"),
        10: dict(name="right_wrist", id=10, color=[204, 51, 51], type="upper", swap="left_wrist"),
        11: dict(name="left_hip", id=11, color=[204, 51, 51], type="lower", swap="right_hip"),
        12: dict(name="right_hip", id=12, color=[204, 51, 51], type="lower", swap="left_hip"),
        13: dict(name="left_knee", id=13, color=[204, 51, 51], type="lower", swap="right_knee"),
        14: dict(name="right_knee", id=14, color=[204, 51, 51], type="lower", swap="left_knee"),
        15: dict(name="left_ankle", id=15, color=[204, 51, 51], type="lower", swap="right_ankle"),
        16: dict(name="right_ankle", id=16, color=[204, 51, 51], type="lower", swap="left_ankle"),
        17: dict(name="head", id=17, color=[204, 51, 51], type="upper", swap=""),
        18: dict(name="neck", id=18, color=[36, 93, 48], type="upper", swap=""),
        19: dict(name="hip", id=19, color=[35, 161, 210], type="lower", swap=""),
        20: dict(name="left_big_toe", id=20, color=[204, 51, 51], type="lower", swap="right_big_toe"),
        21: dict(name="right_big_toe", id=21, color=[204, 51, 51], type="lower", swap="left_big_toe"),
        22: dict(name="left_small_toe", id=22, color=[204, 51, 51], type="lower", swap="right_small_toe"),
        23: dict(name="right_small_toe", id=23, color=[204, 51, 51], type="lower", swap="left_small_toe"),
        24: dict(name="left_heel", id=24, color=[204, 51, 51], type="lower", swap="right_heel"),
        25: dict(name="right_heel", id=25, color=[204, 51, 51], type="lower", swap="left_heel"),
        26: dict(name="spine_01", id=26, color=[248, 144, 171], type="upper", swap=""),
        27: dict(name="spine_02", id=27, color=[248, 144, 171], type="upper", swap=""),
        28: dict(name="spine_03", id=28, color=[248, 144, 171], type="upper", swap=""),
        29: dict(name="spine_04", id=29, color=[194, 134, 91], type="upper", swap=""),
        30: dict(name="spine_05", id=30, color=[194, 134, 91], type="upper", swap=""),
        31: dict(name="left_latissimus", id=31, color=[204, 51, 51], type="upper", swap="right_latissimus"),
        32: dict(name="right_latissimus", id=32, color=[204, 51, 51], type="upper", swap="left_latissimus"),
        33: dict(name="left_clavicle", id=33, color=[204, 51, 51], type="upper", swap="right_clavicle"),
        34: dict(name="right_clavicle", id=34, color=[204, 51, 51], type="upper", swap="left_clavicle"),
        35: dict(name="neck_02", id=35, color=[36, 93, 48], type="upper", swap=""),
        36: dict(name="neck_03", id=36, color=[36, 93, 48], type="upper", swap=""),
    },
    skeleton_info={
        0: dict(link=("left_ankle", "left_knee"), id=0, color=[0, 255, 0]),
        1: dict(link=("left_knee", "left_hip"), id=1, color=[0, 255, 0]),
        2: dict(link=("left_hip", "hip"), id=2, color=[0, 255, 0]),
        3: dict(link=("right_ankle", "right_knee"), id=3, color=[255, 128, 0]),
        4: dict(link=("right_knee", "right_hip"), id=4, color=[255, 128, 0]),
        5: dict(link=("right_hip", "hip"), id=5, color=[255, 128, 0]),
        6: dict(link=("head", "neck_03"), id=6, color=[51, 153, 255]),
        7: dict(link=("neck_03", "neck_02"), id=7, color=[51, 153, 255]),
        8: dict(link=("neck_02", "neck"), id=8, color=[51, 153, 255]),
        9: dict(link=("neck", "spine_05"), id=9, color=[51, 153, 255]),
        10: dict(link=("spine_05", "spine_04"), id=10, color=[51, 153, 255]),
        11: dict(link=("spine_04", "spine_03"), id=11, color=[51, 153, 255]),
        12: dict(link=("spine_03", "spine_02"), id=12, color=[51, 153, 255]),
        13: dict(link=("spine_02", "spine_01"), id=13, color=[51, 153, 255]),
        14: dict(link=("spine_01", "hip"), id=14, color=[51, 153, 255]),
        18: dict(link=("left_shoulder", "left_elbow"), id=18, color=[0, 255, 0]),
        19: dict(link=("left_elbow", "left_wrist"), id=19, color=[0, 255, 0]),
        20: dict(link=("left_shoulder", "right_shoulder"), id=20, color=[255, 128, 0]),
        21: dict(link=("right_shoulder", "right_elbow"), id=21, color=[255, 128, 0]),
        22: dict(link=("right_elbow", "right_wrist"), id=22, color=[255, 128, 0]),
        23: dict(link=("left_eye", "right_eye"), id=23, color=[51, 153, 255]),
        24: dict(link=("head", "nose"), id=24, color=[51, 153, 255]),
        25: dict(link=("nose", "left_eye"), id=25, color=[51, 153, 255]),
        26: dict(link=("nose", "right_eye"), id=26, color=[51, 153, 255]),
        27: dict(link=("left_eye", "left_ear"), id=27, color=[51, 153, 255]),
        28: dict(link=("right_eye", "right_ear"), id=28, color=[51, 153, 255]),
        29: dict(link=("left_ear", "left_shoulder"), id=29, color=[51, 153, 255]),
        30: dict(link=("right_ear", "right_shoulder"), id=30, color=[51, 153, 255]),
        31: dict(link=("left_ankle", "left_big_toe"), id=31, color=[0, 255, 0]),
        32: dict(link=("left_ankle", "left_small_toe"), id=32, color=[0, 255, 0]),
        33: dict(link=("left_ankle", "left_heel"), id=33, color=[0, 255, 0]),
        34: dict(link=("right_ankle", "right_big_toe"), id=34, color=[255, 128, 0]),
        35: dict(link=("right_ankle", "right_small_toe"), id=35, color=[255, 128, 0]),
        36: dict(link=("right_ankle", "right_heel"), id=36, color=[255, 128, 0]),
    },
    # the joint_weights is modified by MMPose Team
    joint_weights=[
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        1.2,
        1.2,
        1.5,
        1.5,
        1.0,
        1.0,
        1.2,
        1.2,
        1.5,
        1.5,
    ]
    + [1.0, 1.5, 1.5]  # Head Top, Neck Base, Hip
    + [1.2] * 6  # Feet
    + [1.2, 1.2, 1.2, 1.2, 1.2]  # Spine_01 to Spine_05
    + [1.0, 1.0]  # Latissimus Dorsi
    + [1.2, 1.2]  # Clavicles
    + [1.5, 1.5],  # Neck_02, Neck_03
    # 'https://github.com/Fang-Haoshu/Halpe-FullBody/blob/master/'
    # 'HalpeCOCOAPI/PythonAPI/halpecocotools/cocoeval.py#L245'
    sigmas=[
        # COCO
        0.026,
        0.025,
        0.025,
        0.035,
        0.035,
        0.079,
        0.079,
        0.072,
        0.072,
        0.062,
        0.062,
        0.107,
        0.107,
        0.087,
        0.087,
        0.089,
        0.089,
        # Central Axix (Halpe/COCO-Wholebody)
        0.026,  # head = head top
        0.04,  # neck = neck base / C7, increased from 0.026 to 0.04
        0.07,  # hip = sacrum midpoint, increased from 0.066 to 0.07
        # Feet
        0.079,
        0.079,
        0.079,
        0.079,
        0.079,
        0.079,
        # Spine
        0.10,  # spine_01 = L5, As with L3, slight ambiguity remains.
        0.10,  # spine_02 = L3, Lower back landmarks tend to be less defined, so a similar value is used.
        0.10,  # spine_03 = T12/L1, The transition zone is less distinct, warranting a higher uncertainty.
        0.08,  # spine_04 = T8, Similar reasoning as T3.
        0.08,  # spine_05 = T3, Increased ambiguity from curvature and soft tissue leads to a moderate sigma.
        # Latissimus Dorsi
        # Rationale: Given the diffuse boundaries of these muscles, a higher sigma is appropriate
        # to capture the inherent annotation variability.
        0.12,
        0.12,
        # Clavicles
        # Rationale: These bony landmarks are usually well defined but can vary with shoulder
        # pose; 0.07 reflects moderate confidence slightly lower than the shoulder joints (0.079).
        0.07,
        0.07,
        # Neck
        0.07,  # neck_02 = C3/C4
        0.06,  # neck_03 = C1
    ],
)



import numpy as np
from OneEuroFilter import OneEuroFilter


class KeypointSmoothing:
    def __init__(
        self,
        num_keypoints,
        freq: float,
        mincutoff: float = 1.0,
        beta: float = 0.0,
        dcutoff: float = 1.0,
    ):
        """Initialize the PoseFiltering class.

        Args:
            num_keypoints (int): Number of keypoints to filter.
            freq (float): An estimate of the frequency in Hz of the signal (> 0), if timestamps are not available.
            mincutoff (float, optional): Min cutoff frequency in Hz (> 0). Lower values allow to remove more jitter.
            beta (float, optional): Parameter to reduce latency (> 0).
            dcutoff (float, optional): Used to filter the derivates. 1 Hz by default. Change this parameter if you know what you are doing.

        Raises:
            ValueError: If one of the parameters is not >0
        """
        kwargs = dict(
            freq=freq,
            mincutoff=mincutoff,
            beta=beta,
            dcutoff=dcutoff,
        )
        self.filters = [
            {
                "filter_x": OneEuroFilter(**kwargs),
                "filter_y": OneEuroFilter(**kwargs),
            }
            for _ in range(num_keypoints)
        ]

    def filter_keypoint(self, point, filters):
        """Filter a single keypoint.

        Args:
            point (np.array): The 3D keypoint to filter. Shape: (3,)
            filters (dict): Filters for x and y coordinates.

        Returns:
            filtered_point (np.array): The filtered 3D keypoint. Shape: (3,)
        """
        filtered_x = filters["filter_x"](point[0])
        filtered_y = filters["filter_y"](point[1])
        return np.array([filtered_x, filtered_y])

    def __call__(self, keypoints):
        """Filter all keypoints in a pose.

        Only smooth the first person.

        Args:
            keypoints_3d (np.array): The 3D keypoints to filter. Shape: (K, 2)
            timestamp (float): The timestamp in seconds.

        Returns:
            filtered_keypoints (np.array): The filtered 3D keypoints. Shape: (num_keypoints, 3)
        """
        return np.array(
            [
                self.filter_keypoint(kpt, filters)
                for kpt, filters in zip(keypoints, self.filters)
            ]
        )



from .tools.base_solution import BasePoseSolution
from .metainfo import metainfo


class SpinePoseEstimator(BasePoseSolution):
    """
    SpinePose: Body + Spine pose estimation using SpineTrack keypoints.

    Combines HALPE-26 keypoints with additional spine keypoints obtained
    from an auxiliary spine pose model.
    """

    MODE = {
        "xlarge": {
            "det": "https://download.openmmlab.com/mmpose/v1/projects/rtmposev1/onnx_sdk/yolox_x_8xb8-300e_humanart-a39d44ed.zip",
            "det_input_size": (640, 640),
            "pose": "https://huggingface.co/saifkhichi96/spinepose/resolve/main/spinepose-x_32xb128-10e_spinetrack-384x288.onnx",
            "pose_input_size": (288, 384),
        },
        "large": {
            "det": "https://download.openmmlab.com/mmpose/v1/projects/rtmposev1/onnx_sdk/yolox_x_8xb8-300e_humanart-a39d44ed.zip",
            "det_input_size": (640, 640),
            "pose": "https://huggingface.co/saifkhichi96/spinepose/resolve/main/spinepose-l_32xb256-10e_spinetrack-256x192.onnx",
            "pose_input_size": (192, 256),
        },
        "medium": {
            "det": "https://download.openmmlab.com/mmpose/v1/projects/rtmposev1/onnx_sdk/yolox_m_8xb8-300e_humanart-c2c7a14a.zip",
            "det_input_size": (640, 640),
            "pose": "https://huggingface.co/saifkhichi96/spinepose/resolve/main/spinepose-m_32xb256-10e_spinetrack-256x192.onnx",
            "pose_input_size": (192, 256),
        },
        "small": {
            "det": "https://download.openmmlab.com/mmpose/v1/projects/rtmposev1/onnx_sdk/yolox_tiny_8xb8-300e_humanart-6f3252f9.zip",
            "det_input_size": (416, 416),
            "pose": "https://huggingface.co/saifkhichi96/spinepose/resolve/main/spinepose-s_32xb256-10e_spinetrack-256x192.onnx",
            "pose_input_size": (192, 256),
        },
    }

    def __init__(
        self,
        mode: str = "large",
        backend="onnxruntime",
        device: str = "auto",
    ):
        super().__init__(
            metainfo,
            self.MODE,
            mode=mode,
            backend=backend,
            device=device,
        )

    def postprocess(self, keypoints, scores):
        keypoints, scores = self._smooth_spine(keypoints, scores)

        # Hide latissimus dorsi and clavicle keypoints
        # These keypoints are not used in the final output
        hidden_ids = [31, 32, 33, 34]
        scores[:, hidden_ids] = 0
        return keypoints, scores

    def _smooth_spine(self, keypoints, scores):
        """Smooth spine keypoints based on domain-specific rule."""
        spine_ids = [36, 35, 18, 30, 29, 28, 27, 26, 19]
        spine_keypoints = keypoints[:, spine_ids]
        spine_scores = scores[:, spine_ids]

        # Smooth by averaging consecutive points
        for i in range(1, len(spine_keypoints[0]) - 1):
            spine_keypoints[:, i] = (
                spine_keypoints[:, i - 1] + spine_keypoints[:, i + 1]
            ) / 2

        # Replace in global keypoints
        keypoints[:, spine_ids] = spine_keypoints
        scores[:, spine_ids] = spine_scores

        return keypoints, scores



import warnings
from typing import Tuple

import numpy as np

from .tools.smoothing import KeypointSmoothing


def compute_iou(bboxA, bboxB):
    """Compute the Intersection over Union (IoU) between two boxes .

    Args:
        bboxA (list): The first bbox info (left, top, right, bottom, score).
        bboxB (list): The second bbox info (left, top, right, bottom, score).

    Returns:
        float: The IoU value.
    """

    x1 = max(bboxA[0], bboxB[0])
    y1 = max(bboxA[1], bboxB[1])
    x2 = min(bboxA[2], bboxB[2])
    y2 = min(bboxA[3], bboxB[3])

    inter_area = max(0, x2 - x1) * max(0, y2 - y1)

    bboxA_area = (bboxA[2] - bboxA[0]) * (bboxA[3] - bboxA[1])
    bboxB_area = (bboxB[2] - bboxB[0]) * (bboxB[3] - bboxB[1])
    union_area = float(bboxA_area + bboxB_area - inter_area)
    if union_area == 0:
        union_area = 1e-5
        warnings.warn("union_area=0 is unexpected")

    return inter_area / union_area


def pose_to_bbox(keypoints: np.ndarray, expansion: float = 1.25) -> np.ndarray:
    """Get bounding box from keypoints.

    Args:
        keypoints (np.ndarray): Keypoints of person.
        expansion (float): Expansion ratio of bounding box.

    Returns:
        np.ndarray: Bounding box of person.
    """
    x = keypoints[:, 0]
    y = keypoints[:, 1]
    bbox = np.array([x.min(), y.min(), x.max(), y.max()])
    center = np.array([bbox[0] + bbox[2], bbox[1] + bbox[3]]) / 2
    return np.concatenate(
        [
            center - (center - bbox[:2]) * expansion,
            center + (bbox[2:] - center) * expansion,
        ]
    )


class PoseTracker:
    """
    Multi-frame pose tracker that adds temporal consistency by associating detections across frames.

    This class delegates single-frame pose estimation to a BasePoseSolution instance,
    and handles tracking (e.g. assigning track IDs and reusing bounding boxes across frames).
    """

    MIN_AREA = 1000

    def __init__(
        self,
        solution: type,
        mode: str = "large",
        backend: str = "onnxruntime",
        device: str = "auto",
        det_frequency: int = 1,
        max_detections: int = 10,
        # Tracking parameters
        tracking: bool = True,
        tracking_thr: float = 0.3,
        # Smoother parameters
        smoothing: bool = False,
        smoothing_freq: float = 30.0,  # Default frequency of the input data (e.g., 30 FPS video)
        smoothing_mincutoff: float = 0.1,  # Lower cutoff for smoothing (higher = less smoothing)
        smoothing_beta: float = 0.1,  # Speed coefficient (higher = more dynamic adaptation)
        smoothing_dcutoff: float = 1.0,  # Derivative cutoff frequency
    ):
        """
        Args:
            solution (BasePoseSolution): An instance of the pose estimation solution.
            det_frequency (int): Frequency of running detection (e.g. every N frames).
            max_detections (int): Maximum number of detections to consider.
            tracking (bool): If True, tracking is enabled.
            tracking_thr (float): IoU threshold for associating bounding boxes.
        """
        self.solution = solution(
            mode=mode,
            backend=backend,
            device=device,
        )
        self.det_frequency = det_frequency
        self.max_detections = max_detections
        self.smoothing = smoothing
        self.smoothing_cfg = dict(
            num_keypoints=self.solution.num_keypoints,
            freq=smoothing_freq,
            mincutoff=smoothing_mincutoff,
            beta=smoothing_beta,
            dcutoff=smoothing_dcutoff,
        )
        self.filters = {}
        self.tracking = tracking or smoothing
        self.tracking_thr = tracking_thr
        self.reset()

    def reset(self):
        """Reset the tracking state."""
        self.frame_cnt = 0
        self.next_id = 0
        self.bboxes_last_frame = []
        self.track_ids_last_frame = []

    def visualize(self, image: np.ndarray, keypoints: np.ndarray, scores: np.ndarray):
        return self.solution.visualize(image, keypoints, scores)

    def __call__(self, image: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Process a single frame: run pose estimation and update tracking.

        Returns:
            keypoints (np.ndarray): The estimated keypoints.
            scores (np.ndarray): The associated confidence scores.
        """
        # Determine bounding boxes using detection (if available) or reuse from last frame
        if self.solution.det_model:
            if self.frame_cnt % self.det_frequency == 0:
                bboxes = self.solution.detect(image)
            else:
                bboxes = self.bboxes_last_frame
        else:
            bboxes = None  # For solutions that don't use detection

        # Run pose estimation (detection + pose + postprocessing)
        keypoints, scores = self.solution.estimate(image, bboxes)

        if not self.tracking:
            # Without tracking, simply compute bounding boxes from keypoints
            bboxes_current_frame = [pose_to_bbox(kpts) for kpts in keypoints]
        else:
            # With tracking, assign track IDs based on IoU matching
            if not self.track_ids_last_frame:
                # Initialize track IDs for the first frame
                self.track_ids_last_frame = (
                    list(range(len(bboxes))) if bboxes is not None else []
                )
                self.next_id = len(self.track_ids_last_frame)

            bboxes_current_frame = []
            new_track_ids = []
            for kpts in keypoints:
                bbox = pose_to_bbox(kpts)
                track_id, _ = self.track_by_iou(bbox)
                if track_id >= 0:
                    new_track_ids.append(track_id)
                    bboxes_current_frame.append(bbox)

            self.track_ids_last_frame = new_track_ids

        # Smooth keypoints if enabled
        if self.smoothing:
            # Map each detection to its track ID, then create or update the smoothing filter
            for i, (kpts, track_id) in enumerate(
                zip(keypoints, self.track_ids_last_frame)
            ):
                if track_id < 0:
                    # Skip smoothing for untracked bboxes
                    continue
                if track_id not in self.filters:
                    # Create a new smoother for this track
                    self.filters[track_id] = KeypointSmoothing(**self.smoothing_cfg)
                # Update the filter with the current keypoints
                smoothed_kpts = self.filters[track_id](kpts)
                keypoints[i] = smoothed_kpts

            # (Optional) Remove filters for track IDs that disappeared this frame
            current_tracks = set(self.track_ids_last_frame)
            disappeared = [tid for tid in self.filters if tid not in current_tracks]
            for tid in disappeared:
                del self.filters[tid]

        # Save state for the next frame and increment frame counter
        self.bboxes_last_frame = bboxes_current_frame
        self.frame_cnt += 1

        return keypoints, scores

    def track_by_iou(self, bbox):
        """
        Greedily assign a track id based on IoU between the current bbox and those from the previous frame.

        Args:
            bbox (list): Current bounding box [left, top, right, bottom].

        Returns:
            tuple: (track_id, max_iou) where track_id is the assigned id (or -1 if no match) and max_iou is the IoU score.
        """
        area = (bbox[2] - bbox[0]) * (bbox[3] - bbox[1])
        max_iou = -1
        max_index = -1
        for index, prev_bbox in enumerate(self.bboxes_last_frame):
            iou = compute_iou(bbox, prev_bbox)
            if iou > max_iou:
                max_iou = iou
                max_index = index

        if max_iou > self.tracking_thr:
            # Match found: reuse the corresponding track ID
            track_id = self.track_ids_last_frame.pop(max_index)
            self.bboxes_last_frame.pop(
                max_index
            )  # remove matched bbox to avoid duplicate matches
        elif area >= self.MIN_AREA:
            # No good match and bbox is large: assign a new track ID
            track_id = self.next_id
            self.next_id += 1
        else:
            # Bbox is too small to track
            track_id = -1

        return track_id, max_iou



import argparse
import os
import warnings

import cv2

from spinepose.pose_estimator import SpinePoseEstimator
from spinepose.pose_tracker import PoseTracker
from spinepose._version import __version__


def imshow(img, title="Image"):
    h, w = img.shape[:2]
    scale = 1024 / max(h, w)
    new_h, new_w = int(h * scale), int(w * scale)
    img = cv2.resize(img, (new_w, new_h))
    cv2.imshow(title, img)


def infer_image(
    input_path,
    mode="medium",
    spine_only=False,
    output_path=None,
):
    model = SpinePoseEstimator(mode)

    img = cv2.imread(input_path)
    keypoints, scores = model(img)

    if spine_only and len(scores) > 0:
        spine_ids = [36, 35, 18, 30, 29, 28, 27, 26, 19]
        non_spine_ids = list(set(range(len(scores[0]))) - set(spine_ids))
        scores[:, non_spine_ids] = 0

    # Show the keypoints on the image
    vis = model.visualize(img, keypoints, scores)
    imshow(vis, "SpinePose Image Inference")
    cv2.waitKey(0)
    cv2.destroyAllWindows()

    # Save the image if requested
    if output_path is not None:
        vis_rgb = cv2.cvtColor(vis, cv2.COLOR_BGR2RGB)
        cv2.imwrite(output_path, vis_rgb)


def infer_video(
    input_path,
    mode="medium",
    spine_only=False,
    use_smoothing=True,
    output_path=None,
):
    cap = cv2.VideoCapture(input_path)
    if not cap.isOpened():
        raise ValueError(f"Cannot open video file {input_path}")

    fps = cap.get(cv2.CAP_PROP_FPS)

    pose_tracker = PoseTracker(
        SpinePoseEstimator,
        mode=mode,
        smoothing=use_smoothing,
        smoothing_freq=fps,
    )

    writer = None
    if output_path is not None:
        try:
            import imageio

            writer = imageio.get_writer(output_path, fps=int(fps))
        except ImportError:
            warnings.warn(
                "Please run `pip install imageio[ffmpeg]` to enable video saving."
                " The video will not be saved.",
                UserWarning,
            )
            writer = None

    while True:
        ret, img = cap.read()
        if not ret:
            break

        keypoints, scores = pose_tracker(img)

        if spine_only and len(scores) > 0:
            spine_ids = [36, 35, 18, 30, 29, 28, 27, 26, 19]
            non_spine_ids = list(set(range(len(scores[0]))) - set(spine_ids))
            scores[:, non_spine_ids] = 0

        vis = pose_tracker.visualize(img, keypoints, scores)

        # Display the result
        imshow(vis, "SpinePose Video Inference")
        if cv2.waitKey(1) & 0xFF == ord("q"):
            break

        # Save the frame if requested
        if writer is not None:
            vis_rgb = cv2.cvtColor(vis, cv2.COLOR_BGR2RGB)
            writer.append_data(vis_rgb)

    cap.release()
    cv2.destroyAllWindows()
    if writer is not None:
        writer.close()


def exists(filepath):
    """Check if the file exists."""
    return os.path.isfile(filepath)


def is_valid(filepath, formats):
    """Check if the file has a valid format."""
    _, ext = os.path.splitext(filepath)
    return ext.lower() in formats


def is_image(filename):
    """Check if the file is an image."""
    img_exts = [".jpg", ".jpeg", ".png", ".bmp"]
    return exists(filename) and is_valid(filename, img_exts)


def is_video(filename):
    """Check if the file is a video."""
    video_exts = [".mp4", ".avi", ".mov", ".mkv"]
    return exists(filename) and is_valid(filename, video_exts)


def main():
    parser = argparse.ArgumentParser(description="SpinePose Inference")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--version",
        "-V",
        action="store_true",
        help="Print the version and exit."
    )
    group.add_argument(
        "--input_path",
        "-i",
        type=str,
        help="Path to the input image or video",
    )
    parser.add_argument(
        "--output_path",
        "-o",
        type=str,
        default=None,
        help="Path to save the output image or video",
    )
    parser.add_argument(
        "--mode",
        "-m",
        choices=["xlarge", "large", "medium", "small"],
        default="medium",
        help="Model size. Choose from: xlarge, large, medium, small (default: medium)",
    )
    parser.add_argument(
        "--nosmooth",
        action="store_false",
        help="Disable keypoint smoothing for video inference (default: enabled)",
    )
    parser.add_argument(
        "--spine_only",
        action="store_true",
        help="Only show spine keypoints (default: show all keypoints)",
    )
    args = parser.parse_args()

    if args.version:
        print(f"SpinePose {__version__}")
        return

    # Check if the input path is a valid image or video
    if is_image(args.input_path):
        infer_image(
            args.input_path,
            args.mode,
            spine_only=args.spine_only,
            output_path=args.output_path,
        )
    elif is_video(args.input_path):
        infer_video(
            args.input_path,
            args.mode,
            spine_only=args.spine_only,
            use_smoothing=args.nosmooth,
            output_path=args.output_path,
        )


if __name__ == "__main__":
    main()



import logging
from typing import Tuple

import numpy as np

from .object_detection import YOLOX
from .pose_estimation import RTMPose
from .utils.multithreading import concurrent_forloop
from .visualization import draw_skeleton


def get_device() -> Tuple[str, str]:
    """
    Get the device for running the model.

    Returns:
        A tuple containing the backend and device.
    """
    try:
        import onnxruntime as ort
        import torch

        available_providers = ort.get_available_providers()

        if torch.cuda.is_available():
            if "CUDAExecutionProvider" in available_providers:
                device, backend = "cuda", "onnxruntime"
                logging.info("Using ONNXRuntime backend with CUDA.")
            elif "ROCMExecutionProvider" in available_providers:
                device, backend = "rocm", "onnxruntime"
                logging.info("Using ONNXRuntime backend with ROCM.")
            else:
                raise RuntimeError("No suitable GPU execution provider found.")
        elif (
            "MPSExecutionProvider" in available_providers
            or "CoreMLExecutionProvider" in available_providers
        ):
            device, backend = "mps", "onnxruntime"
            logging.info("Using ONNXRuntime backend with MPS/CoreML.")
        else:
            raise RuntimeError("No suitable GPU execution provider found.")
    except Exception as e:
        logging.warning(f"Error while checking GPU availability: {e}")

        # Fallback to CPU with OpenVINO
        device, backend = "cpu", "openvino"
        logging.info("Falling back to OpenVINO backend with CPU.")

    return backend, device


class BasePoseSolution:
    """
    Single-frame pose estimation solution.

    This class is responsible for:
      - Running the detection model (if available) to generate bounding boxes.
      - Running the pose estimation model on provided bounding boxes.
      - Optionally postprocessing the output keypoints and scores.

    The __call__ method accepts an image and, optionally, precomputed bounding boxes.
    If no bounding boxes are provided, detection is performed automatically.
    """

    def __init__(
        self,
        metainfo: dict,
        config: dict,
        mode: str = "performance",
        backend: str = "onnxruntime",
        device: str = "auto",
    ):
        self.metainfo = metainfo
        self.num_keypoints = len(metainfo["keypoint_info"])

        mode_config = config.get(mode)
        if mode_config is None:
            logging.warning(
                f"Mode '{mode}' is not supported by {self.__class__.__name__}. Falling back to 'lightweight' mode."
            )
            mode_config = config.get("lightweight")
            if mode_config is None:
                raise ValueError(
                    f"No supported mode found for {self.__class__.__name__}."
                )

        # Set the device and backend
        if device == "auto":
            backend, device = get_device()

        self.backend = backend
        self.device = device

        # Initialize detection and pose models
        self.det_model = YOLOX(
            mode_config["det"],
            model_input_size=mode_config["det_input_size"],
            backend=backend,
            device=device,
        )
        self.pose_model = RTMPose(
            mode_config["pose"],
            model_input_size=mode_config["pose_input_size"],
            backend=backend,
            device=device,
        )

    def detect(self, image: np.ndarray) -> np.ndarray:
        """Run detection to get bounding boxes from the image."""
        return self.det_model(image)

    def estimate(
        self, image: np.ndarray, bboxes: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Run pose estimation on the image using the given bounding boxes.

        Returns:
            A tuple (keypoints, scores) after postprocessing.
        """
        # Process each bounding box concurrently
        results = concurrent_forloop(
            lambda bbox: self.pose_model(image, bboxes=[bbox]),
            bboxes,
        )

        if len(results) == 0:
            # No bounding boxes detected
            return np.zeros((0, self.num_keypoints, 3)), np.zeros(
                (0, self.num_keypoints)
            )

        # Concatenate results
        keypoints, scores = zip(*results)
        keypoints = np.concatenate(keypoints, axis=0)
        scores = np.concatenate(scores, axis=0)

        # Postprocess the results
        return self.postprocess(keypoints, scores)

    def postprocess(
        self, keypoints: np.ndarray, scores: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Optionally postprocess the keypoints and scores.

        Override this method if additional postprocessing is needed.
        """
        return keypoints, scores

    def visualize(
        self, image: np.ndarray, keypoints: np.ndarray, scores: np.ndarray
    ) -> np.ndarray:
        """
        Visualize the keypoints and scores on the image.

        Override this method if custom visualization is needed.
        """
        scale = image.shape[1] / 800
        radius = int(4 * scale)
        line_width = int(2 * scale)
        return draw_skeleton(
            image,
            keypoints,
            scores,
            self.metainfo,
            radius=radius,
            line_width=line_width,
        )

    def __call__(
        self, image: np.ndarray, bboxes: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Run the full single-frame pipeline.

        If `bboxes` is provided, it is used directly; otherwise, detection is performed.
        """
        if bboxes is None:
            bboxes = self.detect(image)
        return self.estimate(image, bboxes)



# spinepose/_version.py

import os

# This assumes that the VERSION file is in the parent directory
# relative to where this file lives. Adjust the path if needed.
VERSION_FILE = os.path.join(os.path.dirname(__file__), "VERSION")

with open(VERSION_FILE, "r", encoding="utf-8") as f:
    __version__ = f.read().strip()



from typing import Tuple

import numpy as np


def get_simcc_maximum(
    simcc_x: np.ndarray, simcc_y: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Get maximum response location and value from simcc representations.

    Note:
        instance number: N
        num_keypoints: K
        heatmap height: H
        heatmap width: W

    Args:
        simcc_x (np.ndarray): x-axis SimCC in shape (K, Wx) or (N, K, Wx)
        simcc_y (np.ndarray): y-axis SimCC in shape (K, Wy) or (N, K, Wy)

    Returns:
        tuple:
        - locs (np.ndarray): locations of maximum heatmap responses in shape
            (K, 2) or (N, K, 2)
        - vals (np.ndarray): values of maximum heatmap responses in shape
            (K,) or (N, K)
    """
    N, K, Wx = simcc_x.shape
    simcc_x = simcc_x.reshape(N * K, -1)
    simcc_y = simcc_y.reshape(N * K, -1)

    # get maximum value locations
    x_locs = np.argmax(simcc_x, axis=1)
    y_locs = np.argmax(simcc_y, axis=1)
    locs = np.stack((x_locs, y_locs), axis=-1).astype(np.float32)
    max_val_x = np.amax(simcc_x, axis=1)
    max_val_y = np.amax(simcc_y, axis=1)

    # get maximum value across x and y axis
    # mask = max_val_x > max_val_y
    # max_val_x[mask] = max_val_y[mask]
    vals = 0.5 * (max_val_x + max_val_y)
    locs[vals <= 0.0] = -1

    # reshape
    locs = locs.reshape(N, K, 2)
    vals = vals.reshape(N, K)

    return locs, vals



import logging
import os
from abc import ABCMeta, abstractmethod
from typing import Any

import cv2
import numpy as np

from .utils.file import download_checkpoint


def check_mps_support():
    try:
        import onnxruntime

        providers = onnxruntime.get_available_providers()
        return (
            "MPSExecutionProvider" in providers
            or "CoreMLExecutionProvider" in providers
        )
    except ImportError:
        return False


RTMLIB_SETTINGS = {
    "opencv": {
        "cpu": (cv2.dnn.DNN_BACKEND_OPENCV, cv2.dnn.DNN_TARGET_CPU),
        # You need to manually build OpenCV through cmake
        "cuda": (cv2.dnn.DNN_BACKEND_CUDA, cv2.dnn.DNN_TARGET_CUDA),
    },
    "onnxruntime": {
        "cpu": "CPUExecutionProvider",
        "cuda": "CUDAExecutionProvider",
        "rocm": "ROCMExecutionProvider",
        "mps": "CoreMLExecutionProvider"
        if check_mps_support()
        else "CPUExecutionProvider",
    },
}


class BaseTool(metaclass=ABCMeta):
    def __init__(
        self,
        onnx_model: str = None,
        model_input_size: tuple = None,
        mean: tuple = None,
        std: tuple = None,
        backend: str = "opencv",
        device: str = "cpu",
    ):
        if not os.path.exists(onnx_model):
            onnx_model = download_checkpoint(onnx_model)

        if backend == "opencv":
            try:
                providers = RTMLIB_SETTINGS[backend][device]

                session = cv2.dnn.readNetFromONNX(onnx_model)
                session.setPreferableBackend(providers[0])
                session.setPreferableTarget(providers[1])
                self.session = session
            except Exception:
                raise RuntimeError(
                    "This model is not supported by OpenCV"
                    " backend, please use `pip install"
                    " onnxruntime` or `pip install"
                    " onnxruntime-gpu` to install onnxruntime"
                    " backend. Then specify `backend=onnxruntime`."
                )  # noqa

        elif backend == "onnxruntime":
            import onnxruntime as ort

            providers = RTMLIB_SETTINGS[backend][device]

            self.session = ort.InferenceSession(
                path_or_bytes=onnx_model, providers=[providers]
            )

        elif backend == "openvino":
            from openvino.runtime import Core

            core = Core()
            model_onnx = core.read_model(model=onnx_model)

            if device != "cpu":
                logging.warning(
                    "OpenVINO only supports CPU backend, automatically"
                    " switched to CPU backend."
                )

            self.compiled_model = core.compile_model(
                model=model_onnx,
                device_name="CPU",
                config={"PERFORMANCE_HINT": "LATENCY"},
            )
            self.input_layer = self.compiled_model.input(0)
            self.output_layer0 = self.compiled_model.output(0)
            self.output_layer1 = self.compiled_model.output(1)

        else:
            raise NotImplementedError

        logging.info(f"load {onnx_model} with {backend} backend")

        self.onnx_model = onnx_model
        self.model_input_size = model_input_size
        self.mean = mean
        self.std = std
        self.backend = backend
        self.device = device

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Any:
        """Implement the actual function here."""
        raise NotImplementedError

    def inference(self, img: np.ndarray):
        """Inference model.

        Args:
            img (np.ndarray): Input image in shape.

        Returns:
            outputs (np.ndarray): Output of RTMPose model.
        """
        # build input to (1, 3, H, W)
        img = img.transpose(2, 0, 1)
        img = np.ascontiguousarray(img, dtype=np.float32)
        input = img[None, :, :, :]

        # run model
        if self.backend == "opencv":
            outNames = self.session.getUnconnectedOutLayersNames()
            self.session.setInput(input)
            outputs = self.session.forward(outNames)
        elif self.backend == "onnxruntime":
            sess_input = {self.session.get_inputs()[0].name: input}
            sess_output = []
            for out in self.session.get_outputs():
                sess_output.append(out.name)

            outputs = self.session.run(sess_output, sess_input)
        elif self.backend == "openvino":
            results = self.compiled_model(input)
            output0 = results[self.output_layer0]
            output1 = results[self.output_layer1]
            outputs = [output0, output1]

        return outputs



from typing import Tuple

import cv2
import numpy as np


def bbox_xyxy2cs(
    bbox: np.ndarray, padding: float = 1.0
) -> Tuple[np.ndarray, np.ndarray]:
    """Transform the bbox format from (x,y,w,h) into (center, scale)

    Args:
        bbox (ndarray): Bounding box(es) in shape (4,) or (n, 4), formatted
            as (left, top, right, bottom)
        padding (float): BBox padding factor that will be multilied to scale.
            Default: 1.0

    Returns:
        tuple: A tuple containing center and scale.
        - np.ndarray[float32]: Center (x, y) of the bbox in shape (2,) or
            (n, 2)
        - np.ndarray[float32]: Scale (w, h) of the bbox in shape (2,) or
            (n, 2)
    """
    # convert single bbox from (4, ) to (1, 4)
    dim = bbox.ndim
    if dim == 1:
        bbox = bbox[None, :]

    # get bbox center and scale
    x1, y1, x2, y2 = np.hsplit(bbox, [1, 2, 3])
    center = np.hstack([x1 + x2, y1 + y2]) * 0.5
    scale = np.hstack([x2 - x1, y2 - y1]) * padding

    if dim == 1:
        center = center[0]
        scale = scale[0]

    return center, scale


def _rotate_point(pt: np.ndarray, angle_rad: float) -> np.ndarray:
    """Rotate a point by an angle.

    Args:
        pt (np.ndarray): 2D point coordinates (x, y) in shape (2, )
        angle_rad (float): rotation angle in radian

    Returns:
        np.ndarray: Rotated point in shape (2, )
    """
    sn, cs = np.sin(angle_rad), np.cos(angle_rad)
    rot_mat = np.array([[cs, -sn], [sn, cs]])
    return rot_mat @ pt


def _get_3rd_point(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """To calculate the affine matrix, three pairs of points are required. This
    function is used to get the 3rd point, given 2D points a & b.

    The 3rd point is defined by rotating vector `a - b` by 90 degrees
    anticlockwise, using b as the rotation center.

    Args:
        a (np.ndarray): The 1st point (x,y) in shape (2, )
        b (np.ndarray): The 2nd point (x,y) in shape (2, )

    Returns:
        np.ndarray: The 3rd point.
    """
    direction = a - b
    return b + np.r_[-direction[1], direction[0]]


def get_warp_matrix(
    center: np.ndarray,
    scale: np.ndarray,
    rot: float,
    output_size: Tuple[int, int],
    shift: Tuple[float, float] = (0.0, 0.0),
    inv: bool = False,
) -> np.ndarray:
    """Calculate the affine transformation matrix that can warp the bbox area
    in the input image to the output size.

    Args:
        center (np.ndarray[2, ]): Center of the bounding box (x, y).
        scale (np.ndarray[2, ]): Scale of the bounding box
            wrt [width, height].
        rot (float): Rotation angle (degree).
        output_size (np.ndarray[2, ] | list(2,)): Size of the
            destination heatmaps.
        shift (0-100%): Shift translation ratio wrt the width/height.
            Default (0., 0.).
        inv (bool): Option to inverse the affine transform direction.
            (inv=False: src->dst or inv=True: dst->src)

    Returns:
        np.ndarray: A 2x3 transformation matrix
    """

    shift = np.array(shift)
    src_w = scale[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    # compute transformation matrix
    rot_rad = np.deg2rad(rot)
    src_dir = _rotate_point(np.array([0.0, src_w * -0.5]), rot_rad)
    dst_dir = np.array([0.0, dst_w * -0.5])

    # get four corners of the src rectangle in the original image
    src = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale * shift
    src[1, :] = center + src_dir + scale * shift
    src[2, :] = _get_3rd_point(src[0, :], src[1, :])

    # get four corners of the dst rectangle in the input image
    dst = np.zeros((3, 2), dtype=np.float32)
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]

    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir
    dst[2, :] = _get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        warp_mat = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        warp_mat = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return warp_mat


def top_down_affine(
    input_size: dict, bbox_scale: dict, bbox_center: dict, img: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Get the bbox image as the model input by affine transform.

    Args:
        input_size (dict): The input size of the model.
        bbox_scale (dict): The bbox scale of the img.
        bbox_center (dict): The bbox center of the img.
        img (np.ndarray): The original image.

    Returns:
        tuple: A tuple containing center and scale.
        - np.ndarray[float32]: img after affine transform.
        - np.ndarray[float32]: bbox scale after affine transform.
    """
    w, h = input_size
    warp_size = (int(w), int(h))

    # reshape bbox to fixed aspect ratio
    aspect_ratio = w / h
    b_w, b_h = np.hsplit(bbox_scale, [1])
    bbox_scale = np.where(
        b_w > b_h * aspect_ratio,
        np.hstack([b_w, b_w / aspect_ratio]),
        np.hstack([b_h * aspect_ratio, b_h]),
    )

    # get the affine matrix
    center = bbox_center
    scale = bbox_scale
    rot = 0
    warp_mat = get_warp_matrix(center, scale, rot, output_size=(w, h))

    # do affine transform
    img = cv2.warpAffine(img, warp_mat, warp_size, flags=cv2.INTER_LINEAR)

    return img, bbox_scale



from typing import List, Tuple

import numpy as np

from ..base_tool import BaseTool
from .post_processings import get_simcc_maximum
from .pre_processings import bbox_xyxy2cs, top_down_affine


class RTMPose(BaseTool):
    def __init__(
        self,
        onnx_model: str,
        model_input_size: tuple = (288, 384),
        mean: tuple = (123.675, 116.28, 103.53),
        std: tuple = (58.395, 57.12, 57.375),
        backend: str = "onnxruntime",
        device: str = "cpu",
    ):
        super().__init__(onnx_model, model_input_size, mean, std, backend, device)

    def __call__(self, image: np.ndarray, bboxes: list = []):
        if len(bboxes) == 0:
            bboxes = [[0, 0, image.shape[1], image.shape[0]]]

        keypoints, scores = [], []
        for bbox in bboxes:
            img, center, scale = self.preprocess(image, bbox)
            outputs = self.inference(img)
            kpts, score = self.postprocess(outputs, center, scale)

            keypoints.append(kpts)
            scores.append(score)

        keypoints = np.concatenate(keypoints, axis=0)
        scores = np.concatenate(scores, axis=0)

        return keypoints, scores

    def preprocess(self, img: np.ndarray, bbox: list):
        """Do preprocessing for RTMPose model inference.

        Args:
            img (np.ndarray): Input image in shape.
            bbox (list):  xyxy-format bounding box of target.

        Returns:
            tuple:
            - resized_img (np.ndarray): Preprocessed image.
            - center (np.ndarray): Center of image.
            - scale (np.ndarray): Scale of image.
        """
        bbox = np.array(bbox)

        # get center and scale
        center, scale = bbox_xyxy2cs(bbox, padding=1.25)

        # do affine transformation
        resized_img, scale = top_down_affine(self.model_input_size, scale, center, img)
        # normalize image
        if self.mean is not None:
            self.mean = np.array(self.mean)
            self.std = np.array(self.std)
            resized_img = (resized_img - self.mean) / self.std

        return resized_img, center, scale

    def postprocess(
        self,
        outputs: List[np.ndarray],
        center: Tuple[int, int],
        scale: Tuple[int, int],
        simcc_split_ratio: float = 2.0,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Postprocess for RTMPose model output.

        Args:
            outputs (np.ndarray): Output of RTMPose model.
            model_input_size (tuple): RTMPose model Input image size.
            center (tuple): Center of bbox in shape (x, y).
            scale (tuple): Scale of bbox in shape (w, h).
            simcc_split_ratio (float): Split ratio of simcc.

        Returns:
            tuple:
            - keypoints (np.ndarray): Rescaled keypoints.
            - scores (np.ndarray): Model predict scores.
        """
        # decode simcc
        simcc_x, simcc_y = outputs
        locs, scores = get_simcc_maximum(simcc_x, simcc_y)
        keypoints = locs / simcc_split_ratio

        # rescale keypoints
        keypoints = keypoints / self.model_input_size * scale
        keypoints = keypoints + center - scale / 2

        return keypoints, scores



# Code modified from https://github.com/IDEA-Research/DWPose/blob/opencv_onnx/ControlNet-v1-1-nightly/annotator/dwpose/cv_ox_det.py  # noqa
from typing import List, Tuple

import cv2
import numpy as np

from ..base_tool import BaseTool
from .post_processings import multiclass_nms


class YOLOX(BaseTool):
    def __init__(
        self,
        onnx_model: str,
        model_input_size: tuple = (640, 640),
        nms_thr=0.45,
        score_thr=0.7,
        backend: str = "onnxruntime",
        device: str = "cpu",
    ):
        super().__init__(onnx_model, model_input_size, backend=backend, device=device)
        self.nms_thr = nms_thr
        self.score_thr = score_thr

    def __call__(self, image: np.ndarray):
        image, ratio = self.preprocess(image)
        outputs = self.inference(image)[0]
        return self.postprocess(outputs, ratio)

    def preprocess(self, img: np.ndarray):
        """Do preprocessing for RTMPose model inference.

        Args:
            img (np.ndarray): Input image in shape.

        Returns:
            tuple:
            - resized_img (np.ndarray): Preprocessed image.
            - center (np.ndarray): Center of image.
            - scale (np.ndarray): Scale of image.
        """
        if len(img.shape) == 3:
            padded_img = (
                np.ones(
                    (self.model_input_size[0], self.model_input_size[1], 3),
                    dtype=np.uint8,
                )
                * 114
            )
        else:
            padded_img = np.ones(self.model_input_size, dtype=np.uint8) * 114

        ratio = min(
            self.model_input_size[0] / img.shape[0],
            self.model_input_size[1] / img.shape[1],
        )
        resized_img = cv2.resize(
            img,
            (int(img.shape[1] * ratio), int(img.shape[0] * ratio)),
            interpolation=cv2.INTER_LINEAR,
        ).astype(np.uint8)
        padded_shape = (int(img.shape[0] * ratio), int(img.shape[1] * ratio))
        padded_img[: padded_shape[0], : padded_shape[1]] = resized_img

        return padded_img, ratio

    def postprocess(
        self,
        outputs: List[np.ndarray],
        ratio: float = 1.0,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Do postprocessing for RTMPose model inference.

        Args:
            outputs (List[np.ndarray]): Outputs of RTMPose model.
            ratio (float): Ratio of preprocessing.

        Returns:
            tuple:
            - final_boxes (np.ndarray): Final bounding boxes.
            - final_scores (np.ndarray): Final scores.
        """

        if outputs.shape[-1] == 4:
            # onnx without nms module

            grids = []
            expanded_strides = []
            strides = [8, 16, 32]

            hsizes = [self.model_input_size[0] // stride for stride in strides]
            wsizes = [self.model_input_size[1] // stride for stride in strides]

            for hsize, wsize, stride in zip(hsizes, wsizes, strides):
                xv, yv = np.meshgrid(np.arange(wsize), np.arange(hsize))
                grid = np.stack((xv, yv), 2).reshape(1, -1, 2)
                grids.append(grid)
                shape = grid.shape[:2]
                expanded_strides.append(np.full((*shape, 1), stride))

            grids = np.concatenate(grids, 1)
            expanded_strides = np.concatenate(expanded_strides, 1)
            outputs[..., :2] = (outputs[..., :2] + grids) * expanded_strides
            outputs[..., 2:4] = np.exp(outputs[..., 2:4]) * expanded_strides

            predictions = outputs[0]
            boxes = predictions[:, :4]
            scores = predictions[:, 4:5] * predictions[:, 5:]

            boxes_xyxy = np.ones_like(boxes)
            boxes_xyxy[:, 0] = boxes[:, 0] - boxes[:, 2] / 2.0
            boxes_xyxy[:, 1] = boxes[:, 1] - boxes[:, 3] / 2.0
            boxes_xyxy[:, 2] = boxes[:, 0] + boxes[:, 2] / 2.0
            boxes_xyxy[:, 3] = boxes[:, 1] + boxes[:, 3] / 2.0
            boxes_xyxy /= ratio
            dets, keep = multiclass_nms(
                boxes_xyxy, scores, nms_thr=self.nms_thr, score_thr=self.score_thr
            )
            if dets is not None:
                pack_dets = (dets[:, :4], dets[:, 4], dets[:, 5])
                final_boxes, final_scores, final_cls_inds = pack_dets
                isscore = final_scores > 0.3
                iscat = final_cls_inds == 0
                isbbox = [i and j for (i, j) in zip(isscore, iscat)]
                final_boxes = final_boxes[isbbox]

        elif outputs.shape[-1] == 5:
            # onnx contains nms module

            pack_dets = (outputs[0, :, :4], outputs[0, :, 4])
            final_boxes, final_scores = pack_dets
            final_boxes /= ratio
            isscore = final_scores > 0.3
            isbbox = list(isscore)
            final_boxes = final_boxes[isbbox]

        return final_boxes
